import asyncio
from time import time

from pytest import approx
import pytest
import asyncio_for_robotics as afor

async def test_soft_wait_for_coro():
    res = await afor.soft_wait_for(asyncio.sleep(1), 0.001)
    assert isinstance(res, TimeoutError), "Should timeout"

async def test_soft_wait_for_fut():
    fut = asyncio.create_task(asyncio.sleep(1))
    res = await afor.soft_wait_for(fut, 0.001)
    assert isinstance(res, TimeoutError), "Should timeout"

async def test_soft_timeout():
    fut = asyncio.create_task(asyncio.sleep(1))
    res = 'oh nyo'
    async with afor.soft_timeout(0.001):
        await asyncio.sleep(1)
        res = 'did not timeout'
    assert res == 'oh nyo', "Should timeout"

async def test_soft_timeout_return_not_timeout():
    async with afor.soft_timeout(0.001) as timeouted:
        await asyncio.sleep(1)
    assert timeouted() == True, "yielded function should return True when timeout occurs"

async def test_soft_timeout_return_timeout():
    async with afor.soft_timeout(1) as timeouted:
        await asyncio.sleep(0.001)
    assert timeouted() == False, "yielded function should return False when timeout doe not occur"

async def test_rate():
    async with afor.soft_timeout(1) as timeouted:
        count = 0
        start = time()
        async for call_time in afor.Rate(100).listen_reliable():
            count += 1
            if count >= 10:
                break
        end = time()
        dt = end - start
        assert dt == approx(0.1, abs=0.1)

    if timeouted() == True:
        pytest.fail(f"test took too long")



import asyncio
import logging
import subprocess
import sys
from typing import Any, AsyncGenerator, Callable, Generator

import pytest
import pytest_asyncio
from base_tests import (
    test_freshness,
    test_listen_one_by_one,
    test_listen_too_fast,
    test_loop_cancellation,
    test_reliable_extremely_fast,
    test_reliable_one_by_one,
    test_reliable_too_fast,
    test_wait_cancellation,
    test_wait_for_value,
    test_wait_new,
    test_wait_next,
)

import asyncio_for_robotics.textio as afor
from asyncio_for_robotics.core._logger import setup_logger

is_win = (
    sys.version_info[0] == 3
    and sys.version_info[1] >= 8
    and sys.platform.startswith("win")
)

pytestmark = pytest.mark.skipif(is_win, reason="Requires a special EvenLoop")

setup_logger(debug_path="tests")
logger = logging.getLogger("asyncio_for_robotics.test")


@pytest.fixture
def session() -> Generator[subprocess.Popen[str], Any, Any]:
    logger.info("Starting process")
    proc = subprocess.Popen(
        "cmd.exe" if sys.platform.startswith("win") else "bash",
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        text=True,
    )
    yield proc
    logger.info("Closing process")
    proc.terminate()
    proc.wait()


@pytest.fixture
def pub(session: subprocess.Popen[str]) -> Generator[Callable[[str], None], Any, Any]:
    def write_in_proc(input: str) -> None:
        assert session.stdin is not None
        res = session.stdin.write(f"""echo "{input}"\n""")
        session.stdin.flush()

    yield write_in_proc


@pytest.fixture
async def sub(session: subprocess.Popen[str]) -> AsyncGenerator[afor.Sub[str], Any]:
    assert session.stdout is not None
    s = afor.from_proc_stdout(session)
    yield s
    s.close()



import logging
from typing import Callable, Dict, Generic, Optional, Type, TypeVar, Union

from rclpy.qos import QoSProfile
from rclpy.subscription import Subscription

from ..core.sub import BaseSub
from .session import BaseSession, auto_session
from .utils import QOS_DEFAULT, TopicInfo, _MsgType

try:
    from typing import Self  # Python 3.11+
except ImportError:
    from typing_extensions import Self  # Python 3.10

logger = logging.getLogger(__name__)


class Sub(BaseSub[_MsgType]):
    def __init__(
        self,
        msg_type: type[_MsgType],
        topic: str,
        qos_profile: QoSProfile = QOS_DEFAULT,
        session: Optional[BaseSession] = None,
    ) -> None:
        """
        Implementation of a asyncio ROS2 subscriber.

        Refere to the base class (BaseSub) for details.

        Args:
            msg_type: The type of ROS messages the subscription will subscribe to.
            topic: The name of the topic the subscription will subscribe to.
            qos: A QoSProfile to apply to the subscription.
            session: The ROS 2 node to use. If not provided, will use
                auto_session to create/get one.
        """
        self.session: BaseSession = self._resolve_session(session)
        self.topic_info = TopicInfo(topic=topic, msg_type=msg_type, qos=qos_profile)
        self.sub = self._resolve_sub(self.topic_info)
        super().__init__()

    @classmethod
    def from_info(
        cls,
        topic_info: TopicInfo,
        session: Optional[BaseSession] = None,
    ) -> Self:
        """Creates the sub from a topic info directly.

        TypeHinting doesn't work with that so... I don't like it
        """
        return cls(**topic_info.as_kwarg(), session=session)

    @property
    def name(self) -> str:
        try:
            return f"ROS2-{self.sub.topic_name}"
        except:
            return f"ROS2-{self.topic_info.topic}"

    def _resolve_session(self, session: Optional[BaseSession]) -> BaseSession:
        """Called at __init__ to get the Node.

        Usefull to overide in a child class and change the Node behavior.
        """
        return auto_session(session)

    def _resolve_sub(self, topic_info: TopicInfo) -> Subscription:
        """Called at __init__ to create the subscriber.

        Usefull to overide in a child class and change the Subscription behavior.
        """
        with self.session.lock() as node:
            return node.create_subscription(
                **topic_info.as_kwarg(),
                callback=self.callback_for_sub,
            )

    def callback_for_sub(self, msg: _MsgType):
        """Callback of the ROS 2 Subscriber.

        This might be executed in another thread.

        Args:
            sample: incoming message
        """
        try:
            healty = self.input_data(msg)
            if not healty:
                self.session._node.destroy_subscription(self.sub)
        except Exception as e:
            logger.error(e)

    def close(self):
        super().close()
        with self.session.lock() as node:
            if not node.executor.context.ok():
                return
            node.destroy_subscription(self.sub)



"""
Example ROS 2 publishers using asyncio integration.

This example demonstrates how to:
- Safely create and destroy publishers using `auto_session().lock()`.
- Publish messages asynchronously at regular intervals with non-blocking
  `asyncio.sleep`.
- Run multiple publishers concurrently and cancel them dynamically.
- Ensure cleanup of publishers when tasks are cancelled or errors occur,
  which is especially useful for testing scenarios with short-lived
  publishers and subscribers.

The script alternates between running:
  - Publisher #1 alone.
  - Publisher #1 and Publisher #2 together.
  - Only Publisher #2.

Run this script using `python3 -m asyncio_for_robotics.example.ros2_double_talker`
Run this script side by side with `python3 -m asyncio_for_robotics.example.ros2_double_listener`
"""
import asyncio
from contextlib import suppress

import rclpy
from std_msgs.msg import String

from asyncio_for_robotics.ros2.session import auto_session
from asyncio_for_robotics.ros2.utils import TopicInfo

TOPIC1 = TopicInfo(msg_type=String, topic="example/talker1")
TOPIC2 = TopicInfo(msg_type=String, topic="example/talker2")

async def pub1_loop():
    # create the publisher safely
    with auto_session().lock() as node:
        pub = node.create_publisher(TOPIC1.msg_type, TOPIC1.topic, TOPIC1.qos)
        node.get_clock(...)
        print("Pub #1 safely created")

    try:
        count = 0
        while 1:  # This loop is not very precise and can drift
            data = f"[Hello world! timestamp: {count}s]"
            count += 1
            print(f"Pub #1 sending: {data}")
            pub.publish(String(data=data)) # sends data (lock is not necessary)
            await asyncio.sleep(1) # non-blocking sleep
    finally:
        # if error or anything occurs, publisher is cleaned up
        # This is not necessary in most application. 
        # It is mainly usefull for tests where many short lived pub/sub are created
        with auto_session().lock() as node:
            node.destroy_publisher(pub)
            print("Pub #1 safely cleaned up")

async def pub2_loop():
    """Same as pub1_loop"""
    with auto_session().lock() as node:
        pub = node.create_publisher(TOPIC2.msg_type, TOPIC2.topic, TOPIC2.qos)
        print("Pub #2 safely created")

    try:
        count = 0
        while 1:
            data = f"[Hello world! timestamp: {count}s]"
            count += 1
            print(f"Pub #2 sending: {data}")
            pub.publish(String(data=data))
            await asyncio.sleep(1)
    finally:
        with auto_session().lock() as node:
            node.destroy_publisher(pub)
            print("Pub #2 safely cleaned up")


async def main():
    while 1:
        print()
        print("Running pub #1 for 5s")
        task1 = asyncio.create_task(pub1_loop())
        await asyncio.sleep(5)
        print()
        print("Running pub #1 and #2 for 5s")
        task2 = asyncio.create_task(pub2_loop())
        await asyncio.sleep(5)
        print()
        print("stopping pub #1, keep running pub #2 for 5s")
        task1.cancel()
        await asyncio.sleep(5)
        print()
        print("stopping pub #2")
        task2.cancel()
        await asyncio.sleep(2)



if __name__ == "__main__":
    rclpy.init()
    try:
        # suppress, just so we don't flood the terminal on exit
        with suppress(KeyboardInterrupt, asyncio.CancelledError):
            asyncio.run(main()) # starts asyncio executor
    finally:
        # cleanup. `finally` statment always executes
        auto_session().close()
        rclpy.shutdown()




import logging
from contextlib import suppress
from typing import Callable, Dict, Generic, Optional, Type, TypeVar, Union

import zenoh

from ..core.sub import BaseSub
from .session import auto_session

logger = logging.getLogger(__name__)


class Sub(BaseSub[zenoh.Sample]):
    def __init__(
        self,
        key_expr: Union[zenoh.KeyExpr, str],
        session: Optional[zenoh.Session] = None,
    ) -> None:
        """
        asyncio_for_robotics implementation of a Zenoh subscriber.

        Refere to the base class (BaseSub) for details.

        Args:
            key_expr:
            session:
        """
        self.session = self._resolve_session(session)
        self.sub = self._resolve_sub(key_expr)
        self._name = f"zenoh-{self.sub.key_expr}"
        super().__init__()

    @property
    def name(self) -> str:
        return self._name

    def _resolve_session(self, session: Optional[zenoh.Session]) -> zenoh.Session:
        """Called at __init__ to get the Session.

        Usefull to overide in a child class and change the Node behavior.
        """
        return auto_session(session)

    def _resolve_sub(self, key_expr: Union[zenoh.KeyExpr, str]):
        """Called at __init__ to create the subscriber.

        Usefull to overide in a child class and change the Subscription behavior.
        """
        return self.session.declare_subscriber(
            key_expr=key_expr, handler=self.callback_for_sub
        )

    def callback_for_sub(self, sample: zenoh.Sample):
        """Zenoh callback happening on the Zenoh thread"""
        if self._closed.is_set():
            return
        try:
            healty = self.input_data(sample)
            if not healty:
                self.close()
        except Exception as e:
            logger.error(e)

    def close(self):
        """Undeclares the subscriber on the zenoh session"""
        if not self._closed.is_set():
            if not self.session.is_closed():
                self.sub.undeclare()
            else:
                logger.debug("Zenoh session already closed for %s", self.name)
        super().close()



import asyncio
import logging
import threading
import uuid
from abc import ABC, abstractmethod
from contextlib import contextmanager, suppress
from typing import Any, Generator, Optional, Union

from rclpy.executors import MultiThreadedExecutor, SingleThreadedExecutor
from rclpy.node import Node
from rclpy.task import Future as FutureRos

try:
    from typing import Self  # Python 3.11+
except ImportError:
    from typing_extensions import Self  # Python 3.10


logger = logging.getLogger(__name__)


class BaseSession(ABC):
    def __init__(
        self,
        node: Union[None, str, Node] = None,
        executor: Union[None, SingleThreadedExecutor, MultiThreadedExecutor] = None,
    ) -> None:
        """Ros2 node spinning in its own thread.

        .. Critical:
            Use the lock() context to safely interact with the node from the main thread.

        .. Important:
            (I think) You cannot have more than one per python instance.

        Args:
            name: name of the node, None will give it a UUID
        """
        logger.debug("Initializing Threaded ROS Sessions")
        name = f"SessionNode{uuid.uuid4()}".replace("-", "_")
        if isinstance(node, str):
            name = node
        if not isinstance(node, Node):
            node = Node(name)
        self._node: Node = node
        if executor is None:
            executor = GLOBAL_SESSION_DEFAULT_EXEC()
        self._executor: Union[SingleThreadedExecutor, MultiThreadedExecutor] = executor
        self._executor.add_node(self._node)
        self._lock = threading.RLock()

    def set_global_session(self) -> None:
        """Make this object instance the global session used by default with auto_session."""
        set_auto_session(self)

    @abstractmethod
    @contextmanager
    def lock(self) -> Generator[Node, Any, Any]:
        """Context to stop the node from spinning momentarly.

        Use like so:
            ```python
            with session.lock() as node:
                node.create_subscription(...)
            # your code continues safely
            ```
        """
        ...

    @abstractmethod
    def start(self):
        """Starts spinning the ros2 node"""
        ...

    def stop(self):
        """Stops the executor and node"""
        with self.lock():
            self._node.destroy_node()
            self._executor.shutdown()

    def close(self):
        global GLOBAL_SESSION
        self.stop()
        if GLOBAL_SESSION is self:
            GLOBAL_SESSION = None


class ThreadedSession(BaseSession):
    _global_node: Optional[Self] = None

    def __init__(
        self,
        node: Union[None, str, Node] = None,
        executor: Union[None, SingleThreadedExecutor, MultiThreadedExecutor] = None,
    ) -> None:
        """Ros2 node spinning in its own background thread.
        ROS2 Callbacks are therefor short and never-blocking.

        .. Critical:
            Use the lock() context to safely interact with the node from the main thread.

        .. Important:
            (I think) In the current version. You cannot have more than one per
            python instance.
        """
        super().__init__(node, executor)
        self.thread = threading.Thread(target=self._spin_thread, daemon=True)
        self._stop_event = threading.Event()
        self._can_spin_event = threading.Event()
        self._node.create_timer(0.01, self._ros_pause_check)
        logger.debug("Initialized Threaded ROS Sessions")

    @contextmanager
    def lock(self) -> Generator[Node, Any, Any]:
        """This context is required for every operation (aside from start/stop).
        Safely, quickly acquires lock on the whole ros2 thread.

        Use like so:
            ```python
            with session.lock() as node:
                node.create_subscription(...)
            # your code continues safely
            ```

        This will pause every ros2 callbacks and processes of the executor in
        use by this object. Only use it shortly to add something to the node
        (pub/sub, timer...)

        This context pauses the executor spinning, and acquires the _lock around it. To
        be double safe.

        When exiting this context, the internal _lock is freed and executor resumed.
        """
        logger.debug("lock requested")
        was_running = self._can_spin_event.is_set()
        try:
            self._pause()
            with self._lock:
                yield self._node
        finally:
            if was_running:
                self._resume()

    def start(self):
        """Starts spinning the ros2 node in its thread"""
        self._stop_event.clear()
        if not self.thread.is_alive():
            logger.debug("RosNode thread started")
            self._resume()
            self.thread.start()

    def stop(self):
        self._stop_event.set()
        super().stop()
        self.thread.join()
        logger.debug("RosNode thread stoped")

    def _spin_thread(self):
        """(thrd 2) Executes in a second thread to spin the node"""

        def aok():
            if self._executor is None:
                exec = True
            else:
                exec = self._executor.context.ok()
            return exec and not self._stop_event.is_set()

        while aok():
            try:
                self._can_spin_event.wait(timeout=1)
                if not aok():
                    return
            except TimeoutError:
                # checks back on aok every timeout
                continue
            logger.debug("ROS waiting for lock")
            with self._lock:
                logger.debug("ROS has lock, spinning")
                # spins until the pause future is triggered
                self.pause_fut = FutureRos()
                self._executor.spin_until_future_complete(self.pause_fut)
                logger.debug("ROS spinning paused")
            logger.debug("ROS released lock")
        logger.debug("ROS spinning TERMINATED")

    def _ros_pause_check(self):
        """(thrd 2) Timer checking if ros spin should pause"""
        try:
            pause_called = not self._can_spin_event.is_set()
            stop_called = self._stop_event.is_set()
            if stop_called or pause_called:
                if self.pause_fut.done():
                    return
                self.pause_fut.set_result(None)
                logger.debug(
                    f"Rclpy pause set pause_called=%s, stop_called=%s",
                    pause_called,
                    stop_called,
                )
        except Exception as e:
            print(e)
            logger.critical(e)

    def _pause(self):
        self._can_spin_event.clear()

    def _resume(self):
        self._can_spin_event.set()


class SynchronousSession(BaseSession):
    _global_node: Optional[Self] = None

    def __init__(
        self,
        node: Union[None, str, Node] = None,
        executor: Union[None, SingleThreadedExecutor, MultiThreadedExecutor] = None,
    ) -> None:
        """Ros2 node spinning in as periodic asyncio task.

        .. Important:
            Unlike ThreadedSession, lock is not necessary to interact with the node.
        """
        super().__init__(node, executor)
        self.rosloop_task: Optional[asyncio.Task] = None

    async def _spin_periodic(self):
        """Executes in the thread to spin the node"""
        while 1:
            await asyncio.sleep(0)
            self._executor.spin_once(timeout_sec=0.0)

    @contextmanager
    def lock(self) -> Generator[Node, Any, Any]:
        try:
            yield self._node
        finally:
            pass

    def start(self):
        """Starts spinning the ros2 node in a asyncio task"""
        logger.debug("RosNode asyncio started")
        self._event_loop = asyncio.get_event_loop()
        self.rosloop_task = self._event_loop.create_task(self._spin_periodic())

    def stop(self):
        """Stops spinning the ros2 node and joins the thread"""
        super().stop()
        if self.rosloop_task is not None:
            self.rosloop_task.cancel()
        logger.debug("RosNode asyncio stoped")


#: type of executor used by default
GLOBAL_SESSION_DEFAULT_EXEC: Union[
    type[SingleThreadedExecutor], type[MultiThreadedExecutor]
] = SingleThreadedExecutor

#: type of session created by default
GLOBAL_SESSION_DEFAULT_TYPE: type[BaseSession] = ThreadedSession

#: global share session (singleton)
GLOBAL_SESSION: Optional[BaseSession] = None


def set_auto_session(session: Optional[BaseSession] = None) -> None:
    """Set the global shared session instance.

    If called with a session, it replaces the current global session.
    If called with None, the global session is unset (but not close!).

    Args:
        session: ros session to set as default
    """
    global GLOBAL_SESSION
    GLOBAL_SESSION = session


def auto_session(session: Optional[BaseSession] = None) -> BaseSession:
    """Return an active session, returning or creating a global one if needed.

    If a session is passed, it is returned as-is.
    If no session is given and no global session exists,
      a default session is created using the configured default type
      and executor, started, and stored globally.
    If no session is given and a global session exists,
      global session is returned


    Args:
        session: Session potentially empty (None)

    Returns:
        An active session, returning or creating a global one if needed.
    """
    global GLOBAL_SESSION, GLOBAL_SESSION_DEFAULT_EXEC, GLOBAL_SESSION_DEFAULT_TYPE
    if session is not None:
        return session
    if GLOBAL_SESSION is None:
        ses = GLOBAL_SESSION_DEFAULT_TYPE(
            node=None, executor=GLOBAL_SESSION_DEFAULT_EXEC()
        )
        GLOBAL_SESSION = ses
    else:
        ses = GLOBAL_SESSION
    ses.start()
    return ses



import pytest

pytest.importorskip("rclpy")
pytest.importorskip("yaml")

import logging
from typing import Any, AsyncGenerator, Callable, Generator

import rclpy
from rclpy.publisher import Publisher
from rclpy.qos import QoSProfile
from std_msgs.msg import String
from test_ros2_pubsub_thr import (
    TOPIC,
    sub,
    test_freshness,
    test_listen_one_by_one,
    test_listen_too_fast,
    test_loop_cancellation,
    test_reliable_extremely_fast,
    test_reliable_one_by_one,
    test_reliable_too_fast,
    test_wait_cancellation,
    test_wait_for_value,
    test_wait_new,
    test_wait_next,
)

import asyncio_for_robotics.ros2 as aros
from asyncio_for_robotics.ros2.session import SynchronousSession, set_auto_session

logger = logging.getLogger("asyncio_for_robotics.test")


@pytest.fixture(scope="module", autouse=True)
def rclpy_init():
    logger.info("Starting rclpy")
    rclpy.init()
    yield
    logger.info("closing rclpy ")
    rclpy.shutdown()


@pytest.fixture(scope="function", autouse=True)
async def session(rclpy_init) -> AsyncGenerator[aros.BaseSession, Any]:
    logger.info("Starting session")
    set_auto_session(SynchronousSession())
    ses = aros.auto_session()
    yield ses
    logger.info("closing session")
    ses.close()


@pytest.fixture(scope="function")
def pub(session: aros.BaseSession) -> Generator[Callable[[str], None], Any, Any]:
    with session.lock() as node:
        publisher = node.create_publisher(*TOPIC.as_arg())

    def write_in_proc(input: str) -> None:
        publisher.publish(String(data=input))

    yield write_in_proc
    with session.lock() as node:
        node.destroy_publisher(publisher)



from .. import ConverterSub, Rate, soft_timeout, soft_wait_for
from .service import Client, Server
from .session import (
    GLOBAL_SESSION,
    BaseSession,
    SynchronousSession,
    ThreadedSession,
    auto_session,
    set_auto_session,
)
from .sub import Sub
from .utils import QOS_DEFAULT, QOS_TRANSIENT, TopicInfo

__all__ = [
    "ConverterSub",
    "soft_wait_for",
    "soft_timeout",
    "Rate",
    "Server",
    "Client",
    "auto_session",
    "set_auto_session",
    "GLOBAL_SESSION",
    "ThreadedSession",
    "SynchronousSession",
    "BaseSession",
    "Sub",
    "TopicInfo",
    "QOS_TRANSIENT",
    "QOS_DEFAULT",
]



from .core.utils import soft_timeout, soft_wait_for, Rate
from .core.sub import BaseSub, ConverterSub

__all__ = [
    "soft_wait_for",
    "soft_timeout",
    "Rate",
    "BaseSub",
    "ConverterSub",
]



"""
Verbose example showing how to integrate ROS 2 with asyncio_for_robotics.

This script demonstrates:
- Initializing ROS 2 nodes and creating a publisher with `auto_session()`.
- Safely declaring publishers using the session lock to manipulate the node
  running in a background thread.
- Publishing data in the background while running subscriber examples.
- Calling `python_discussion.discuss` to illustrate asyncio-based subscription
  methods (`wait_for_value`, `wait_for_new`, `wait_for_next`, `listen`,
  and `listen_reliable`). See `python_discussion.py` for detailed behavior.
- Properly shutting down both the ROS 2 session and `rclpy` to allow clean
  program exit.
"""
import asyncio
from contextlib import suppress
import rclpy

from rclpy.qos import QoSProfile
from std_msgs.msg import String

from asyncio_for_robotics.ros2.session import auto_session
from asyncio_for_robotics.ros2.sub import Sub
from asyncio_for_robotics.ros2.utils import TopicInfo

from .python_discussion import brint, discuss

TOPIC = TopicInfo(
    msg_type=String,
    topic="example/discussion",
    qos=QoSProfile(
        depth=100,
    ),
)


async def talking_loop():
    print()
    brint(
        f"""asyncio_for_robotics does not provide ros2 publishers, it is out
          of scope. You remain in full control of publishing. You can get the
          ros node to declare a publisher by using `auto_session()`"""
    )
    brint(
        f"""
        Manipulating the node spinning in the background thread is slightly
        different than usual. You need to enter the session.lock() context to
        safely add pub, sub, services ... to the node. This lock will block the
        node for as long as the context is active = as long as your code is
        indented in the `with session.lock() as node:` codeblock.
          """
    )
    print(
f"""Here is how the publisher of this demo is created:

with auto_session().lock() as node:
    pub = node.create_publisher(TOPIC.msg_type, TOPIC.topic, TOPIC.qos)
"""
            )
    with auto_session().lock() as node:
        pub = node.create_publisher(TOPIC.msg_type, TOPIC.topic, TOPIC.qos)
    print(f"ROS 2 started publishing onto {pub.topic_name}")
    try:
        count = 0
        while 1:
            pub.publish(String(data=f"[Hello world! timestamp: {count/10:.1f}s]"))
            count += 1
            await asyncio.sleep(0.1)
    finally:
        with auto_session().lock() as node:
            node.destroy_publisher(pub)


def get_str_from_msg(msg: String):
    return msg.data


async def main():
    background_talker_task = asyncio.create_task(talking_loop())
    await asyncio.sleep(0.001)
    print("\n#####################")
    brint(
        f"""
        From now on, all objects are initialized and the code is the same
        between ros2 and zenoh!
        """
    )
    print("#####################\n")
    sub = Sub(**TOPIC.as_kwarg())
    await discuss(sub, get_str_from_msg)
    background_talker_task.cancel()


if __name__ == "__main__":
    print()
    brint(
        f""" ROS requires `rclpy.init` to be called once per python process
        (then shutdown). Unlike the session, asyncio_for_robotics will
        not call `rclpy.init` automatically.
        """
    )
    brint(
        f""" 
        The concept of session is however also present in
        asyncio_for_robotics.ros2! The session is made of a node and executor
        running in a background thread. You can create your own session
        with your node and your executor, or let asyncio_for_robotics automatically
        handle it with `ros2.auto_session()`. `ros2.auto_session()` will be called automatically every time a session is not provided as argument.
        """
    )
    rclpy.init()
    try:
        # suppress, just so we don't flood the terminal on exit
        with suppress(KeyboardInterrupt, asyncio.CancelledError):
            asyncio.run(main()) # starts asyncio executor
    finally:
        # cleanup. `finally` statment always executes
        print()
        brint(
            f"""
        To finish and let python exit, the session and rclpy needs to be closed. We can
        retrieve and close the one automatically created with `auto_session().close()`
        """
        )
        auto_session().close()
        rclpy.shutdown()

